/*
 * ext_all.c - Dump entire configuration at external interface
 *
 * Written 2001 by Werner Almesberger 
 * Copyright 2001 Network Robots
 */

/*
 * This is experimental code and will change. Probably a lot.
 *
 * Known problems: @@@
 * - doesn't understand concept of default parameters
 * - prints flag parameters with numeric argument (is this a feature ?)
 * - does not support qdiscs shared by multiples classes
 * - gets confused if dumping more than one qdisc tree (ingress/egress) per
 *   session
 * - does not take into account that only part of the dsmark class may be used
 *   for marking
 * - does not take into account that no dsmark class may exist for marking
 *   (in fact, should look at parameters to decide whether a change is
 *   actually requested)
 * - does not report inner qdiscs changing skb->tc_index
 */


#include <inttypes.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "error.h"
#include "param.h"
#include "tree.h"
#include "filter.h"
#include "qdisc.h"
#include "ext_all.h"


int dump_all; /* dump all qdiscs and filters (experimental) */


static const QDISC *_root; /* @@@@ HACK */


/* -------------------- skb->tc_index to class mapping --------------------- */


static void print_class_n(FILE *file,const QDISC *qdisc,uint16_t class,
  int *first)
{
    fprintf(file,"%c%u:%u",*first ? ' ' : ',',(unsigned) qdisc->number,
      (unsigned) class);
    *first = 0;
}


static void print_class(FILE *file,const CLASS *class,int *first)
{
    print_class_n(file,class->parent.qdisc,class->number,first);
}


static void gred_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    const CLASS *class;

    param_get(qdisc->params);
    for (class = qdisc->classes; class; class = class->next)
	if (class->number == (index & 0xf)) {
	    print_class(file,class,&first);
	    return;
	}
    print_class_n(file,qdisc,prm_default_index.v,&first);
}


static void no_filters_at_classes(const CLASS *class)
{
    while (class) {
	if (class->filters)
	    yyerror("dump_all currently does not support filters at classes");
	no_filters_at_classes(class->child);
	class = class->next;
    }
}


static void find_path(FILE *file,const QDISC *qdisc,uint16_t index,int first);


static void filter_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    const ELEMENT *element;
    int shift;

    if (!qdisc->filters) {
	if (qdisc->dsc->flags & QDISC_HAS_CLASSES)
	    yyerrorf("don't know how to select classes of \"%s\"",
	      qdisc->dsc->name);
	return;
    }
    if (qdisc->filters->next)
	yyerrorf("dump_all currently supports only one filter per qdisc");
    if (qdisc->filters->dsc != &tcindex_dsc)
	yyerrorf("dump_all supports only \"tcindex\", not \"%s\"",
	  qdisc->filters->dsc->name);
    param_get(qdisc->filters->params);
    shift = prm_shift.v;
    for (element = qdisc->filters->elements; element; element = element->next) {
	param_get(element->params);
	if (((index >> shift) & (prm_mask.present ? prm_mask.v : 0xffff)) ==
	  element->number) {
	    print_class(file,element->parent.class,&first);
	    if (element->parent.class->qdisc)
		find_path(file,element->parent.class->qdisc,index,first);
	    return;
	}
    }
    yyerror("dump_all found no matching choice in tcindex");
}


static void find_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    no_filters_at_classes(qdisc->classes);
    if (qdisc->dsc == &gred_dsc) gred_path(file,qdisc,index,first);
    else filter_path(file,qdisc,index,first);
}


int dump_all_decision(FILE *file,uint16_t class)
{
    const CLASS *c;

    if (!dump_all) return 0;
    for (c = _root->classes; c; c = c->next)
	if (c->qdisc) {
	    int one = 1;

	    fprintf(file," class");
	    print_class_n(file,_root,class,&one);
	    find_path(file,c->qdisc,class,0);
	    return 1;
	}
    return 0;
}


/* --------------------------- Print parameters ---------------------------- */


/*
 * Print parameters in (more or less) alphabetic order.
 */


static void dump_params(FILE *file,PARAM_DEF *def,PARAM *params)
{
    const PARAM_DSC *req = NULL,*opt = NULL;
    const PARAM_DSC **req_ptr,**opt_ptr;

    param_get(params);
    if (def->required) {
	req_ptr = def->required;
	req = *req_ptr++;
    }
    if (def->optional) {
	opt_ptr = def->optional;
	opt = *opt_ptr++;
    }
    while (req || opt) {
	const PARAM_DSC *p;

	if (!req) {
	    p = opt;
	    opt = *opt_ptr++;
	}
	else {
	    if (!opt || strcmp(req->id,opt->id) < 0) {
		p = req;
		req = *req_ptr++;
	    }
	    else {
		p = opt;
		opt = *opt_ptr++;
	    }
	}
	if (p->type == dt_fnum || p->type == dt_string || p->type == dt_list ||
	  p->type == dt_police) continue;
        if (p->present) fprintf(file," %s %lu",p->id,(unsigned long) p->v);
    }
    
}


/* ------------------- Dumps qdisc and class definitions ------------------- */


static void dump_class(FILE *file,const CLASS *class)
{
    fprintf(file,"class %u =",(unsigned) class->number);
    dump_params(file,class->parent.qdisc->dsc->class_param,class->params);  
    fprintf(file,"\n");
}


static void dump_classes(FILE *file,const CLASS *class)
{
    while (class) {
	dump_class(file,class);
	dump_classes(file,class->child);
	class = class->next;
    }
}


static void dump_this_qdisc(FILE *file,const QDISC *qdisc);


static void dump_qdiscs_at_classes(FILE *file,const CLASS *class)
{
    while (class) {
	if (class->qdisc) dump_this_qdisc(file,class->qdisc);
	dump_qdiscs_at_classes(file,class->child);
	class = class->next;
    }
}


static void dump_this_qdisc(FILE *file,const QDISC *qdisc)
{
    fprintf(file,"qdisc %u = %s",(unsigned) qdisc->number,qdisc->dsc->name);
    dump_params(file,qdisc->dsc->qdisc_param,qdisc->params);  
    fprintf(file,"\n");
    dump_classes(file,qdisc->classes);
    dump_qdiscs_at_classes(file,qdisc->classes);
}


/* ------------------------------------------------------------------------- */


void dump_all_qdisc(FILE *file,const QDISC *root)
{
    if (!root) yyerror("dump_all: no root qdisc");
    _root = root;
    if (root->filters) {
	if (root->filters->next) yyerror("dump_all: only one filter allowed");
	if (root->filters->dsc != &if_dsc)
	    yyerrorf("dump_all: filter must be \"if\", not \"%s\"",
	      root->filters->dsc->name);
    }
    if (root->dsc != &dsmark_dsc)
	yyerrorf("dump_all: root qdisc must be \"dsmark\", not \"%s\"",
	  root->dsc->name);
    dump_this_qdisc(file,root);
}
